// wave.c

#include <windows.h>

#include "wave.h"

#define MAXWAVES 4

struct
{
   LPWAVEFORMATEX lpWaveFormatEx;
   LPSTR lpData;
   DWORD dwBufferLength;
   BOOL inUse;
}
g_waves [MAXWAVES];

static void CALLBACK WaveOutProc (HWAVEOUT hwo, UINT uMsg, DWORD dwInstance,
                                  DWORD dwParam1, DWORD dwParam2);

// =================================================
// void CloseWave (int iWaveID)
//
// Close a specific previously opened wave resource.
//
// Arguments:
//
// iWaveID - wave resource identifier
//
// Return:
//
// none
// =================================================

void CloseWave (int iWaveID)
{
   free (g_waves [iWaveID].lpWaveFormatEx);

   free (g_waves [iWaveID].lpData);

   g_waves [iWaveID].inUse = FALSE;
}

// ===================================================
// int OpenWave (LPSTR szWaveResName)
//
// Open a specific wave resource.
//
// Arguments:
//
// szWaveResName - name of wave resource
//
// Return:
//
// positive integer wave resource identifier (success)
// negative integer error code (failure)
// ===================================================

int OpenWave (LPSTR szWaveResName)
{
   int i;
   HMMIO hmmio;
   HRSRC hrsrc;
   DWORD dwSize;
   LPSTR lpData;
   LPSTR lpWave;
   HGLOBAL hglob;
   MMIOINFO mmioinfo;
   MMCKINFO mmckinfoParentChk;
   MMCKINFO mmckinfoChildChk;
   LPWAVEFORMATEX lpWaveFormatEx;

   for (i = 0; i < MAXWAVES; i++)
        if (!g_waves [i].inUse)
            break;

   if (i == MAXWAVES)
       return ERR_TOO_MANY_WAVES;

   hrsrc = FindResource (GetModuleHandle (NULL), szWaveResName, "WAVE");
   if (hrsrc == NULL)
       return ERR_CANNOT_FIND_RES;

   hglob = LoadResource (GetModuleHandle (NULL), hrsrc);
   if (hglob == NULL)
       return ERR_CANNOT_LOAD_RES;

   dwSize = SizeofResource (GetModuleHandle (NULL), hrsrc);
   if (dwSize == NULL)
       return ERR_CANNOT_SIZE_RES;

   lpWave = LockResource (hglob);
   if (lpWave == NULL)
       return ERR_CANNOT_LOCK_RES;

   ZeroMemory (&mmioinfo, sizeof(MMIOINFO));
   mmioinfo.fccIOProc = FOURCC_MEM;
   mmioinfo.pchBuffer = lpWave;
   mmioinfo.cchBuffer = dwSize;

   if ((hmmio = mmioOpen (NULL, &mmioinfo, MMIO_READ)) == NULL)
       return ERR_CANNOT_OPEN_RES;

   mmckinfoParentChk.fccType = mmioFOURCC('W', 'A', 'V', 'E'); 
   if (mmioDescend (hmmio, (LPMMCKINFO) &mmckinfoParentChk, 0, MMIO_FINDRIFF)) 
   {
       mmioClose (hmmio, 0);
       return ERR_WAVE_EXPECTED;
   }

   mmckinfoChildChk.ckid = mmioFOURCC('f', 'm', 't', ' '); 
   if (mmioDescend (hmmio, &mmckinfoChildChk, &mmckinfoParentChk,
       MMIO_FINDCHUNK))
   {
       mmioClose (hmmio, 0);
       return ERR_FMT_EXPECTED;
   }

   if ((lpWaveFormatEx = (LPWAVEFORMATEX) calloc (mmckinfoChildChk.cksize, 1))
       == NULL)
   {
       mmioClose (hmmio, 0);
       return ERR_INSUF_MEMORY;
   }

   if (mmioRead (hmmio, (HPSTR) lpWaveFormatEx, mmckinfoChildChk.cksize) !=
       (LRESULT) mmckinfoChildChk.cksize)
   {
       free (lpWaveFormatEx);
       mmioClose (hmmio, 0);
       return ERR_BAD_FMT;
   }

   mmioAscend (hmmio, &mmckinfoChildChk, 0); 

   mmckinfoChildChk.ckid = mmioFOURCC('d', 'a', 't', 'a');
   if (mmioDescend (hmmio, &mmckinfoChildChk, &mmckinfoParentChk,
       MMIO_FINDCHUNK))
   {
       free (lpWaveFormatEx);
       mmioClose (hmmio, 0);
       return ERR_BAD_DATA;
   }

   if ((lpData = (char *) malloc (mmckinfoChildChk.cksize)) == NULL)
   {
       free (lpWaveFormatEx);
       mmioClose (hmmio, 0);
       return ERR_INSUF_MEMORY;
   }

   if (mmioRead(hmmio, (HPSTR) lpData, mmckinfoChildChk.cksize) !=
                (long) mmckinfoChildChk.cksize)
   {
       free (lpData);
       free (lpWaveFormatEx);
       mmioClose (hmmio, 0);
       return ERR_CANNOT_READ;
   }

   mmioClose (hmmio, 0);

   g_waves [i].lpWaveFormatEx = lpWaveFormatEx;

   g_waves [i].lpData = lpData;

   g_waves [i].dwBufferLength = mmckinfoChildChk.cksize;

   g_waves [i].inUse = TRUE;

   return i;
}

// =========================================================================
// int PlayWave (int iWaveID)
//
// Play a specific wave sound simultaneously with other playing wave sounds.
//
// Arguments:
//
// iWaveID - wave sound identifier
//
// Return:
//
// zero (success) or negative integer error code (failure)
// =========================================================================

int PlayWave (int iWaveID)
{
   HWAVEOUT hwo;
   LPWAVEHDR lpWaveHdr;

   if ((lpWaveHdr = (LPWAVEHDR) malloc (sizeof(WAVEHDR))) == NULL)
       return ERR_INSUF_MEMORY;

   ZeroMemory (lpWaveHdr, sizeof(WAVEHDR));
   lpWaveHdr->dwBufferLength = g_waves [iWaveID].dwBufferLength;
   lpWaveHdr->lpData = g_waves [iWaveID].lpData;

   if (waveOutOpen (&hwo, WAVE_MAPPER, g_waves [iWaveID].lpWaveFormatEx,
       (DWORD) WaveOutProc, NULL, CALLBACK_FUNCTION) != MMSYSERR_NOERROR)
   {
       free (lpWaveHdr);
       return ERR_CANNOT_OPEN_WAVE;
   }

   if (waveOutPrepareHeader (hwo, lpWaveHdr, sizeof(WAVEHDR)) !=
       MMSYSERR_NOERROR)
   {
       free (lpWaveHdr);
       waveOutClose (hwo);
       return ERR_CANNOT_PREPARE;
   }

   if (waveOutWrite (hwo, lpWaveHdr, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
   {
       free (lpWaveHdr);
       waveOutClose (hwo);
       return ERR_CANNOT_WRITE;
   }

   return 0;
}

// ===========================================================================
// void CALLBACK WaveOutProc (HWAVEOUT hwo, UINT uMsg, DWORD dwInstance,
//                            DWORD dwParam1, DWORD dwParam2)
//
// Wave procedure for processing waveform messages related to the progress of 
// sound playback. Example: WOM_DONE is sent when a wave sound stops playing.
//
// Arguments:
//
// hwo        - waveform device handle (as returned by waveOutOpen())
//
// uMsg       - message identifier (only WOM_DONE is processed below)
//
// dwInstance - user-instance data passed to waveOutOpen() (not used)
//
// dwParam1   - message parameter (pointer to WAVEHDR structure that's
//              associated with wave sound, for WOM_DONE message)
//
// dwParam2   - message parameter (not used by the WOM_DONE message)
//
// Return:
//
// none
// ===========================================================================

#pragma argsused
static void CALLBACK WaveOutProc (HWAVEOUT hwo, UINT uMsg, DWORD dwInstance,
                                  DWORD dwParam1, DWORD dwParam2)
{
   LPWAVEHDR lpWaveHdr;

   switch (uMsg)
   {
      case WOM_DONE:
           lpWaveHdr = (LPWAVEHDR) dwParam1;

           waveOutUnprepareHeader (hwo, lpWaveHdr, sizeof(WAVEHDR));
           waveOutClose (hwo);

           free (lpWaveHdr);
   }
}
